/**
 * USART driver.
 * @author Ryan Hall
 */
component USART {

    // bit names for UCSRnA register
    field RXCn:  8 = 0x80;  // USART receive complete
    field TXCn:  8 = 0x40;  // USART transmit complete
    field UDREn: 8 = 0x20;  // USART data registe empty
    field FEn:   8 = 0x10;  // Frame error : set to 1 when writing to UCSRnA
    field DORn:  8 = 0x08;  // Data over run : set to 0 when writing to UCSRnA
    field UPEn:  8 = 0x04;  // Parity error : set to 0 when writing to UCSRnA
    field U2Xn:  8 = 0x02;  // Double USART transmission speed : set to 0 in sync mode
    field MPCMn: 8 = 0x01;  // Multi-processor communication mode

    // bit names for UCSRnB register
    field RXCIEn: 8 = 0x80;  // enable receive complete interrupt
    field TXCIEn: 8 = 0x40;  // enable transmission complete interrupt
    field UDRIEn: 8 = 0x20;  // enable data register empty interrupt
    field RXENn:  8 = 0x10;  // enable receiver
    field TXENn:  8 = 0x08;  // enable transmitter
    field UCSZn2: 8 = 0x04;  // high end bit of 3 for frame size
    field RXB8n:  8 = 0x02;  // 9th bit of received character
    field TXB8n:  8 = 0x01;  // 9th bit of transmitted character

    // bit names for UCSRnC register
    field UMSELn: 8 = 0x40;  // 0 for async, 1 for sync mode
    field UPMn1:  8 = 0x20;  // Parity mode
    field UPMn0:  8 = 0x10;  // Parity mode
    field USBSn:  8 = 0x08;  // Number of stop bits inserted by transmitter
    field UCSZn1: 8 = 0x04;  // Character size low 2 bits
    field UCSZn0: 8 = 0x02;  // Character size low 2 bits
    field UCPOLn: 8 = 0x01;  // Clock polarity in sync mode

    // Baud Rate Registers
    // UBRRnL : lower 8 bits of baud rate
    // UBRRnH : higher 4 bits of baud rate

    // Data Register
    // UDRn

    method calculateBaud(fosc: int, rate: int, mode: int): int {
	if ( mode == 0 ) return fosc / (16 * rate) - 1;
	if ( mode == 1 ) return fosc / (8 * rate) - 1;
	if ( mode == 2 ) return fosc / (2 * rate) - 1;
	return 0;
    }
}

component USART0 {

    field r_queue : Queue<char> = new Queue<char>(10);  // receive queue
    field t_queue : Queue<char> = new Queue<char>(10);  // transmit queue

    method enable(baud_rate: int, sync_mode : boolean, double_async : boolean) {

        // set character size to 8
        device.UCSR0C = USART.UCSZn1 | USART.UCSZn0;
     
        if (sync_mode) {
            device.UCSR0C = device.UCSR0C | USART.UMSELn;
        } else if (double_async) {
            device.UCSR0A = USART.U2Xn;
        }
      
        // set baud rate
        device.UBRR0H = (baud_rate >> 8) :: 8;
        device.UBRR0L = baud_rate :: 8;
 
        // enable interrupts, receiver and transmitter
        device.UCSR0B = USART.RXENn | USART.TXENn | USART.RXCIEn;
    }

    method disable() {
       // disable transmitter and receiver
       device.UCSR0B = 0x0;
    }

    method transmit(buf : char[], i : int, l : int) : int {
        local index : int = 0;
	local mask = MCU.disable();
        for (index = 0 ; index < l ; index++) {
            if (t_queue.enqueue(buf[i + index])) { 
                if (t_queue.size() > 0) {
                    // turn on data ready interrupts
                    device.UCSR0B = device.UCSR0B | USART.UDRIEn; 
                }
            } else {
		MCU.restore(mask);
                return index;
            }
        }
	MCU.restore(mask);
        return index;    
    }

    method tran_handler() {
        local c = t_queue.dequeue();
        if (t_queue.size() == 0) {
            // turn off data ready interrupts
            device.UCSR0B = device.UCSR0B & ~USART.UDRIEn;
        }
        device.UDR0 = c;
    }

    method receive(buf : char[], i : int, l : int) : int {
        local count : int = 0;
	local mask = MCU.disable();
        while (count < l and r_queue.size() > 0) {
            buf[i + count] = r_queue.dequeue();        
            count++;
        }
	MCU.restore(mask);
        return count;
    }

    method rec_handler() {
        local c : char = (device.UDR0) :: char;
        r_queue.enqueue(c);
    }
}

component USART1 {

    field r_queue : Queue<char> = new Queue<char>(10);  // receive queue
    field t_queue : Queue<char> = new Queue<char>(10);  // transmit queue
     
    method enable(baud_rate: int, sync_mode : boolean, double_async : boolean) {

        // set character size to 8
        device.UCSR1C = USART.UCSZn1 | USART.UCSZn0;
     
        if (sync_mode) {
            device.UCSR1C = device.UCSR1C | USART.UMSELn;
        } else if(double_async) {
            device.UCSR1A = USART.U2Xn;
        }
      
        // set baud rate
        device.UBRR1H = (baud_rate >> 8) :: 8;
        device.UBRR1L = baud_rate :: 8;
 
        // enable interrupts, receiver and transmitter
        device.UCSR1B = USART.RXENn | USART.TXENn | USART.RXCIEn;
    }

    method disable() {
        // disable receiver, transmitter and reciever
        device.UCSR1B = 0x0;
    }

    method transmit(buf : char[], i : int, l : int) : int {
        local index : int = 0;
	local mask = MCU.disable();
        for (index = 0 ; index < l ; index++) {
            if (t_queue.enqueue(buf[i + index])) { 
                if (t_queue.size() > 0) {
                    // turn on data ready interrupts
                   device.UCSR1B = device.UCSR1B | USART.UDRIEn; 
                }
            } else {
		MCU.restore(mask);
                return index;
            }
        }
	MCU.restore(mask);
        return index;    
    }

    method tran_handler() {
        local c = t_queue.dequeue();
        if (t_queue.size() == 0) {
            //turn off data ready interrupts
            device.UCSR1B = device.UCSR1B & ~USART.UDRIEn;
        }
        device.UDR1 = c;
    }

    method receive(buf : char[], i : int, l : int) : int {
        local count : int = 0;
	local mask = MCU.disable();
        while(count < l and r_queue.size() > 0) {
            buf[i + count] = r_queue.dequeue();        
            count++;
        }
	MCU.restore(mask);
        return count;
    }

    method rec_handler() {
        local c : char = (device.UDR1) :: char;
        r_queue.enqueue(c);
    }
}
